#include <future>
#include <iostream>

#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/function.hpp>

#include "ray/common/status.h"

#include "ray/common/common_protocol.h"
#include "ray/object_manager/object_store_notification_manager.h"
#include "ray/util/util.h"

namespace ray {

ObjectStoreNotificationManager::ObjectStoreNotificationManager(
    boost::asio::io_service &io_service, const std::string &store_socket_name)
    : store_client_(),
      length_(0),
      num_adds_processed_(0),
      num_removes_processed_(0),
      socket_(io_service) {
  RAY_ARROW_CHECK_OK(store_client_.Connect(store_socket_name.c_str(), "", 0, 300));

  RAY_ARROW_CHECK_OK(store_client_.Subscribe(&c_socket_));
  boost::system::error_code ec;
#if defined(BOOST_ASIO_HAS_LOCAL_SOCKETS)
  local_stream_protocol sp;
#else  // TODO(mehrdadn): HACK: FIXME: This is just to get things compiling!
  local_stream_protocol sp(AF_UNIX, 0);
#endif
  socket_.assign(sp, c_socket_, ec);
  assert(!ec.value());
  NotificationWait();
}

ObjectStoreNotificationManager::~ObjectStoreNotificationManager() {
  RAY_ARROW_CHECK_OK(store_client_.Disconnect());
}

void ObjectStoreNotificationManager::NotificationWait() {
  boost::asio::async_read(socket_, boost::asio::buffer(&length_, sizeof(length_)),
                          boost::bind(&ObjectStoreNotificationManager::ProcessStoreLength,
                                      this, boost::asio::placeholders::error));
}

void ObjectStoreNotificationManager::ProcessStoreLength(
    const boost::system::error_code &error) {
  notification_.resize(length_);
  if (error) {
    // When shutting down a cluster, it's possible that the plasma store is killed
    // earlier than raylet, in this case we don't want raylet to crash, we instead
    // log an error message and exit.
    RAY_LOG(ERROR) << "Failed to process store length: "
                   << boost_to_ray_status(error).ToString()
                   << ", most likely plasma store is down, raylet will exit";
    // Exit raylet process.
    _exit(kRayletStoreErrorExitCode);
  }
  boost::asio::async_read(
      socket_, boost::asio::buffer(notification_),
      boost::bind(&ObjectStoreNotificationManager::ProcessStoreNotification, this,
                  boost::asio::placeholders::error));
}

void ObjectStoreNotificationManager::ProcessStoreNotification(
    const boost::system::error_code &error) {
  if (error) {
    RAY_LOG(FATAL)
        << "Problem communicating with the object store from raylet, check logs or "
        << "dmesg for previous errors: " << boost_to_ray_status(error).ToString();
  }

  const auto &object_notification =
      flatbuffers::GetRoot<object_manager::protocol::PlasmaNotification>(
          notification_.data());
  for (size_t i = 0; i < object_notification->object_info()->size(); ++i) {
    auto object_info = object_notification->object_info()->Get(i);
    const ObjectID object_id =
        ObjectID::FromPlasmaIdBinary(object_info->object_id()->str());
    if (object_info->is_deletion()) {
      ProcessStoreRemove(object_id);
    } else {
      object_manager::protocol::ObjectInfoT result;
      object_info->UnPackTo(&result);
      ProcessStoreAdd(result);
    }
  }
  NotificationWait();
}

void ObjectStoreNotificationManager::ProcessStoreAdd(
    const object_manager::protocol::ObjectInfoT &object_info) {
  for (auto &handler : add_handlers_) {
    handler(object_info);
  }
  num_adds_processed_++;
}

void ObjectStoreNotificationManager::ProcessStoreRemove(const ObjectID &object_id) {
  for (auto &handler : rem_handlers_) {
    handler(object_id);
  }
  num_removes_processed_++;
}

void ObjectStoreNotificationManager::SubscribeObjAdded(
    std::function<void(const object_manager::protocol::ObjectInfoT &)> callback) {
  add_handlers_.push_back(std::move(callback));
}

void ObjectStoreNotificationManager::SubscribeObjDeleted(
    std::function<void(const ObjectID &)> callback) {
  rem_handlers_.push_back(std::move(callback));
}

std::string ObjectStoreNotificationManager::DebugString() const {
  std::stringstream result;
  result << "ObjectStoreNotificationManager:";
  result << "\n- num adds processed: " << num_adds_processed_;
  result << "\n- num removes processed: " << num_removes_processed_;
  return result.str();
}

}  // namespace ray
